@@ -2,10 +2,11 @@ import logging
 import os
 import sys
 import importlib
+import weakref
 from watchdog.observers import Observer
 from watchdog.events import FileSystemEventHandler
 from openhab_pythonrule_engine.item_registry import ItemRegistry
-from openhab_pythonrule_engine.trigger import Trigger
+from openhab_pythonrule_engine.rule import Rule, ExecutionListener
 from openhab_pythonrule_engine.cron_processor import CronProcessor
 from openhab_pythonrule_engine.item_change_processor import ItemChangeProcessor
 from openhab_pythonrule_engine.loaded_rule_processor import RuleLoadedProcessor
@@ -16,63 +17,68 @@ logging = logging.getLogger(__name__)
 
 class FileSystemListener(FileSystemEventHandler):
 
-    def __init__(self, rule_engine, dir):
-        self.rule_engine = rule_engine
+    def __init__(self, rule_engine_ref: weakref, dir):
+        self.rule_engine_ref = rule_engine_ref
         self.dir = dir
-        logging.info("observing rules directory " + dir)
         self.observer = Observer()
 
+    def __unload_module(self, path: str):
+        rule_engine = self.rule_engine_ref()
+        if rule_engine is not None:
+            rule_engine.unload_module(path)
+
+    def __load_module(self, path: str):
+        rule_engine = self.rule_engine_ref()
+        if rule_engine is not None:
+            rule_engine.load_module(path)
+
     def start(self):
+        logging.info("observing rules directory '" + self.dir + "' started")
         for file in os.scandir(self.dir):
-            self.rule_engine.load_module(file.name)
+            self.__load_module(file.name)
         self.observer.schedule(self, self.dir, recursive=False)
         self.observer.start()
 
     def stop(self):
         self.observer.stop()
-        for file in os.scandir(dir):
-            self.rule_engine.unload_module(file.name)
+        logging.info("observing rules directory '" + self.dir + "' stopped")
 
     def on_moved(self, event):
-        self.rule_engine.unload_module(self.filename(event.src_path))
-        self.rule_engine.load_module(self.filename(event.dest_path))
+        self.__unload_module(self.filename(event.src_path))
+        self.__load_module(self.filename(event.dest_path))
 
     def on_deleted(self, event):
-        self.rule_engine.unload_module(self.filename(event.src_path))
+        self.__unload_module(self.filename(event.src_path))
 
     def on_created(self, event):
-        self.rule_engine.load_module(self.filename(event.src_path))
+        self.__load_module(self.filename(event.src_path))
 
     def on_modified(self, event):
-        self.rule_engine.load_module(self.filename(event.src_path))
+        self.__load_module(self.filename(event.src_path))
 
     def filename(self, path):
         path = path.replace("\\", "/")
         return path[path.rindex("/")+1:]
 
 
-
-class RuleEngine:
+class RuleEngine(ExecutionListener):
 
     def __init__(self, openhab_uri:str, python_rule_directory: str, user: str, pwd: str):
         self.is_running = False
         self.openhab_uri = openhab_uri
         self.__item_registry = ItemRegistry(openhab_uri, user, pwd)
-        self.__processors = [ItemChangeProcessor(openhab_uri, self.__item_registry, self.on_executed),
-                             CronProcessor(self.__item_registry, self.on_executed),
-                             RuleLoadedProcessor(self.__item_registry, self.on_executed)]
-        self.file_system_listener = FileSystemListener(self, python_rule_directory)
+        self.__processors = [ItemChangeProcessor(openhab_uri, self.__item_registry, weakref.ref(self)),
+                             CronProcessor(self.__item_registry, weakref.ref(self)),
+                             RuleLoadedProcessor(self.__item_registry, weakref.ref(self))]
+        self.file_system_listener = FileSystemListener(weakref.ref(self), python_rule_directory)
         self.listeners = set()
 
-    def triggers(self):
-        return [trigger for processor in self.__processors for trigger in processor.triggers]
+    def rules(self):
+        return [rule for processor in self.__processors for rule in processor.rules]
 
-    def on_executed(self, trigger: Trigger, error: Exception):
+    def on_executed(self, rule: Rule, error: Exception):
         self.__notify_listener()
 
-    def __del__(self):
-        self.stop()
-
     def add_listener(self, listener):
         self.listeners.add(listener)
         self.__notify_listener()
@@ -86,16 +92,25 @@ class RuleEngine:
 
     def start(self):
         if not self.is_running:
+            logging.info("starting rule engine...")
             self.is_running = True
             if self.python_rule_directory not in sys.path:
                 sys.path.insert(0, self.python_rule_directory)
             [processor.start() for processor in self.__processors]
             self.file_system_listener.start()
+            logging.info("rule engine started")
+
+    def __del__(self):
+        self.stop()
 
     def stop(self):
+        logging.info("stopping rule engine...")
         self.is_running = False
         self.file_system_listener.stop()
         [processor.stop() for processor in self.__processors]
+        for module in {rule.module for rule in self.rules()}:
+            self.unload_module(module + ".py")
+        logging.info("rule engine stopped")
 
     @property
     def python_rule_directory(self):
@@ -108,7 +123,7 @@ class RuleEngine:
                 msg = None
                 # reload?
                 if modulename in sys.modules:
-                    [processor.remove_triggers(modulename) for processor in self.__processors]
+                    [processor.remove_rules(modulename) for processor in self.__processors]
                     importlib.reload(sys.modules[modulename])
                     msg = "'" + filename + "' reloaded"
                 else:
@@ -126,10 +141,10 @@ class RuleEngine:
             try:
                 modulename = self.__filename_to_modulename(filename)
                 if modulename in sys.modules:
-                    if not silent:
-                        logging.info("\"unloading\" '" + filename + "'")
-                    [processor.remove_triggers(modulename) for processor in self.__processors]
+                    [processor.remove_rules(modulename) for processor in self.__processors]
                     del sys.modules[modulename]
+                    if not silent:
+                        logging.info("'" + filename + "' unloaded")
                 self.__notify_listener()
             except Exception as e:
                 logging.warning("error occurred by unloading " + filename + " " + str(e), e)
